/*         ______   ___    ___
 *        /\  _  \ /\_ \  /\_ \
 *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___
 *         \ \  __ \ \ \ \  \ \ \   /'__`\ /'_ `\/\`'__\/ __`\
 *          \ \ \/\ \ \_\ \_ \_\ \_/\  __//\ \L\ \ \ \//\ \L\ \
 *           \ \_\ \_\/\____\/\____\ \____\ \____ \ \_\\ \____/
 *            \/_/\/_/\/____/\/____/\/____/\/___L\ \/_/ \/___/
 *                                           /\____/
 *                                           \_/__/
 *
 *      Improved screen display API, draft.
 *
 *      See readme.txt for copyright information.
 */


#include "allegro.h"
#include "allegro/internal/aintern.h"

#include "display.h"

AL_DISPLAY *al_main_display = NULL;

/* al_create_display
 * Create a new Allegro display object, return NULL on failure
 * The first display object created becomes the al_main_display
 */
AL_DISPLAY *al_create_display(int driver, int flags, int depth, int w, int h, int v_w, int v_h)
{
   AL_DISPLAY *new_display;
   int c;

   ASSERT(system_driver);
   
   /* It's an error to select more than one update method */
   if ( (flags & AL_UPDATE_ALL) & ( (flags & AL_UPDATE_ALL) - 1) )
      return NULL;

   new_display = malloc(sizeof *new_display);
   
   if (!new_display)
      return NULL;
      
   /* Adjust virtual width and virtual height if needed and allowed */
#ifdef ALLEGRO_VRAM_SINGLE_SURFACE
   if ((v_w==0) && (v_h==0) && (flags&(AL_UPDATE_TRIPLE_BUFFER|AL_UPDATE_PAGE_FLIP))) {
      v_w = w;
      v_h = h * (flags&AL_UPDATE_TRIPLE_BUFFER?3:2);
   }
#endif

   /* Set graphics mode */   
   if (depth>0)
      set_color_depth(depth);
   if (set_gfx_mode(driver, w, h, v_w, v_h) == -1) {
      free(new_display);
      return NULL;
   }
   
   new_display->gfx_driver = gfx_driver;
   new_display->screen = screen;
   new_display->page = NULL;
   new_display->num_pages = 0;
   new_display->active_page = 0;
   new_display->flags = flags;
   new_display->depth = get_color_depth();
   new_display->gfx_capabilities = gfx_capabilities;
   
   /* Select update method */
   switch(new_display->flags & AL_UPDATE_ALL) {
      case AL_UPDATE_TRIPLE_BUFFER:
         /* Try to enable triple buffering */
         if (!(new_display->gfx_capabilities & GFX_CAN_TRIPLE_BUFFER)) {
            enable_triple_buffer();
            new_display->gfx_capabilities = gfx_capabilities;
         }

         if (new_display->gfx_capabilities & GFX_CAN_TRIPLE_BUFFER) {
            new_display->num_pages = 3;
            new_display->page = malloc(new_display->num_pages * sizeof *(new_display->page));

            for (c=0; c<new_display->num_pages; c++)
               new_display->page[c] = create_video_bitmap(w, h);
            
            /* Check for succes */
            if (new_display->page[0] && new_display->page[1] && new_display->page[2]) {
               for (c=0; c<new_display->num_pages; c++)
                  clear_bitmap(new_display->page[c]);
               show_video_bitmap(new_display->page[2]);
            }
            else {
               for (c=0; c<new_display->num_pages; c++)
                  destroy_bitmap(new_display->page[c]);
               free(new_display->page);
               set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
               free(new_display);
               return NULL;
            }
         }
         else {
            set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
            free(new_display);
            return NULL;
         }
         break;

      case AL_UPDATE_PAGE_FLIP:
         new_display->num_pages = 2;
         new_display->page = malloc(new_display->num_pages * sizeof *(new_display->page));

         for (c=0; c<new_display->num_pages; c++)
            new_display->page[c] = create_video_bitmap(w, h);
         
         /* Check for succes */
         if (new_display->page[0] && new_display->page[1]) {
            for (c=0; c<new_display->num_pages; c++)
               clear_bitmap(new_display->page[c]);
            show_video_bitmap(new_display->page[1]);
         }
         else {
            for (c=0; c<new_display->num_pages; c++)
               destroy_bitmap(new_display->page[c]);
            free(new_display->page);
            set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
            free(new_display);
            return NULL;
         }
         break;

      case AL_UPDATE_SYSTEM_BUFFER:
      case AL_UPDATE_DOUBLE_BUFFER:
         new_display->num_pages = 1;
         new_display->page = malloc(new_display->num_pages * sizeof *(new_display->page));

         for (c=0; c<new_display->num_pages; c++)
            if (new_display->flags & AL_UPDATE_SYSTEM_BUFFER)
               new_display->page[c] = create_system_bitmap(w, h);
            else
               new_display->page[c] = create_bitmap(w, h);
         
         /* Check for succes */
         if (new_display->page[0]) {
            for (c=0; c<new_display->num_pages; c++)
               clear_bitmap(new_display->page[c]);
            show_video_bitmap(new_display->page[1]);
         }
         else {
            for (c=0; c<new_display->num_pages; c++)
               destroy_bitmap(new_display->page[c]);
            free(new_display->page);
            set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
            free(new_display);
            return NULL;
         }
         break;
   }
   
   if (!al_main_display)
      al_main_display = new_display;
      
   return new_display;
}

/* al_set_update_method:
 * Change the update method for the selected display. If the method change 
 * fails the current method is either retained or reset to AL_UPDATE_NONE.
 * Returns the method being used after this call.
 */
int al_set_update_method(AL_DISPLAY *display, int method)
{
   int c, w, h;
   ASSERT(display);
   ASSERT(display->gfx_driver);

   /* Requesting multiple modes is not valid */
   if ( (method & (method - 1)) ||  (method & ~AL_UPDATE_ALL) )
      return display->flags & AL_UPDATE_ALL;

   /* Do nothing if the methods are the same */      
   if (method == (display->flags & AL_UPDATE_ALL))
      return display->flags & AL_UPDATE_ALL;

   /* Retain current method if triple buffering was requested, but can't work */
   if ( (method & AL_UPDATE_TRIPLE_BUFFER) && 
        (!(display->gfx_capabilities & GFX_CAN_TRIPLE_BUFFER)))
      return display->flags & AL_UPDATE_ALL;
      
   /* Destroy old setup */
   display->flags &= ~AL_UPDATE_ALL;
   for(c=0; c<display->num_pages; c++)
      destroy_bitmap(display->page[c]);
   free(display->page);
   display->page = NULL;
   display->num_pages = 0;
   display->active_page = 0;

   w = display->gfx_driver->w;
   h = display->gfx_driver->h;

   /* Set new method */
   switch(method) {
      case AL_UPDATE_TRIPLE_BUFFER:
         display->num_pages = 3;
         display->page = malloc(display->num_pages * sizeof *(display->page));

         for (c=0; c<display->num_pages; c++)
            display->page[c] = create_video_bitmap(w, h);
            
         /* Check for succes */
         if (display->page[0] && display->page[1] && display->page[2]) {
            for (c=0; c<display->num_pages; c++)
               clear_bitmap(display->page[c]);
            show_video_bitmap(display->page[2]);
            display->flags |= method;
         }
         else {
            for (c=0; c<display->num_pages; c++)
               destroy_bitmap(display->page[c]);
            display->page = NULL;
            display->num_pages = 0;
         }
         break;

      case AL_UPDATE_PAGE_FLIP:
         display->num_pages = 2;
         display->page = malloc(display->num_pages * sizeof *(display->page));

         for (c=0; c<display->num_pages; c++)
            display->page[c] = create_video_bitmap(w, h);
         
         /* Check for succes */
         if (display->page[0] && display->page[1]) {
            for (c=0; c<display->num_pages; c++)
               clear_bitmap(display->page[c]);
            show_video_bitmap(display->page[1]);
            display->flags |= method;
         }
         else {
            for (c=0; c<display->num_pages; c++)
               destroy_bitmap(display->page[c]);
            display->page = NULL;
            display->num_pages = 0;
         }
         break;

      case AL_UPDATE_SYSTEM_BUFFER:
      case AL_UPDATE_DOUBLE_BUFFER:
         display->num_pages = 1;
         display->page = malloc(display->num_pages * sizeof *(display->page));

         for (c=0; c<display->num_pages; c++)
            if (display->flags & AL_UPDATE_SYSTEM_BUFFER)
               display->page[c] = create_system_bitmap(w, h);
            else
               display->page[c] = create_bitmap(w, h);
         
         /* Check for succes */
         if (display->page[0]) {
            for (c=0; c<display->num_pages; c++)
               clear_bitmap(display->page[c]);
            show_video_bitmap(display->page[1]);
            display->flags |= method;
         }
         else {
            for (c=0; c<display->num_pages; c++)
               destroy_bitmap(display->page[c]);
            display->page = NULL;
            display->num_pages = 0;
         }
         break;
   }

   return display->flags & AL_UPDATE_ALL;
}

/* al_destroy_display:
 * Destroys an Allegro display and returns to text mode
 */
void al_destroy_display(AL_DISPLAY *display)
{
   int c;
   
   ASSERT(system_driver);
   
   if (!display)
      return;
      
   for(c=0; c<display->num_pages; c++)
      destroy_bitmap(display->page[c]);
   free(display->page);
   
   set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
   
   free(display);
}

/* al_flip_display:
 * Flip the front and back buffers
 */
void al_flip_display(AL_DISPLAY *display)
{
   ASSERT(display);

   switch(display->flags & AL_UPDATE_ALL) {
      case AL_UPDATE_TRIPLE_BUFFER:
         while (poll_scroll());
         request_video_bitmap(display->page[display->active_page]);
         
         display->active_page = (display->active_page+1)%display->num_pages;
         return;

      case AL_UPDATE_PAGE_FLIP:
         show_video_bitmap(display->page[display->active_page]);

         display->active_page = (display->active_page+1)%display->num_pages;
         return;

      case AL_UPDATE_SYSTEM_BUFFER:
      case AL_UPDATE_DOUBLE_BUFFER:
         /* Wait for vsync */
         if (!(display->flags&AL_DISABLE_VSYNC))
            vsync();
         blit(display->page[0], display->screen, 0, 0, 0, 0, display->page[0]->w, display->page[0]->h);
         return;
   } /* End of switch */
}

/* al_get_buffer:
 * Get a pointer to the backbuffer
 */
BITMAP *al_get_buffer(const AL_DISPLAY *display)
{
   ASSERT(display);
   
   if (display->flags & AL_UPDATE_ALL)
      return display->page[display->active_page];
      
   return display->screen;
}

/* al_get_update_method:
 * Get the update method being used to update the display
 */
int al_get_update_method(const AL_DISPLAY *display)
{
   ASSERT(system_driver);

   return display->flags & AL_UPDATE_ALL;
}


/* VSync settings */

/* al_enable_vsync:
 * Enable vsync for the display
 */
void al_enable_vsync(AL_DISPLAY *display)
{
   ASSERT(display);
   
   display->flags&=~AL_DISABLE_VSYNC;
}

/* al_disable_vsync:
 * disable vsync for the display
 */
void al_disable_vsync(AL_DISPLAY *display)
{
   ASSERT(display);
   
   display->flags|=AL_DISABLE_VSYNC;
}

/* al_toggle_vsync:
 * Toggle vsync for the display
 */
void al_toggle_vsync(AL_DISPLAY *display)
{
   ASSERT(display);
   
   display->flags^=AL_DISABLE_VSYNC;
}

/* al_vsync_is_enabled:
 * returns TRUE if vsync is enabled on the display
 */
int al_vsync_is_enabled(const AL_DISPLAY *display)
{
   ASSERT(display);
   
   return (display->flags&AL_DISABLE_VSYNC) == 0;
}
